from keras import callbacks
from keras.src import backend
from keras import ops
from keras.src.utils import io_utils


class LRCSVLogger(callbacks.CSVLogger):
    """
    Learning rate is logged alongside metrics and loss
    """
    def on_epoch_end(self, epoch, logs=None):
        try:
            current_lr = self.model.optimizer._get_current_learning_rate()  # returns current learning rate
            logs["learning_rate"] = float(
            backend.convert_to_numpy(current_lr)
        )
        except AttributeError:
            pass
        
        return super().on_epoch_end(epoch, logs)
    

class WeightNormLogger(callbacks.CSVLogger):
    def on_train_begin(self, logs=None):
        super().on_train_begin(logs)
        self.initial_norm = self.get_norm_weights()

    def on_train_batch_end(self, batch, logs=None):
        logs = {}

        # write to logs
        logs["weights_norm"] = self.get_norm_weights()
        logs["relative_change"] = logs["weights_norm"] / self.initial_norm
        return super().on_epoch_end(batch, logs)
    
    def get_norm_weights(self) -> float:
        # get model weights
        weights: list = self.model.get_weights()
        
        # flatten list and get norm over all weights
        flat_norm = [ops.linalg.norm(ops.ravel(w)) for w in weights]
        norm = ops.linalg.norm(flat_norm)
        return float(backend.convert_to_numpy(norm))


class ThresholdEarlyStopping(callbacks.EarlyStopping):
    """
    ThresholdEarlyStopping Callback

    This callback adds a threshold parameter to the EarlyStopping callback.
    If threshold is set, the early stopping will not be active until the monitored loss falls below the threshold.

    Parameters
    ----------
    monitor : str, optional
        Quantity to be monitored. Defaults to `"val_loss"`.
    min_delta : float, optional
        Minimum change in the monitored quantity to qualify as an improvement, 
        i.e. an absolute change of less than min_delta, will count as no improvement. 
        Defaults to `0`.
    patience : int, optional
        Number of epochs with no improvement after which training will be stopped. Defaults to `0`.
    verbose : int, optional {0, 1}
        Verbosity mode. 
        Mode 0 is silent, and mode 1 displays messages when the callback takes an action. 
        Defaults to `0`.
    mode : str, optional {"auto", "min", "max"}
        In `min` mode, training will stop when the quantity monitored has stopped decreasing; 
        in `"max"` mode it will stop when the quantity monitored has stopped increasing; 
        in `"auto"` mode, the direction is automatically inferred from the name of the monitored quantity. 
        Defaults to `"auto"`.
    threshold : float, optional
        Threshold value for the monitored quantity.
        If not `None`, the callback will not be active until the monitored quantity does not undercut the threshold.
        Defaults to `None`.
    baseline : float, optional
        Baseline value for the monitored quantity. 
        If not `None`, training will stop if the model doesn't show improvement over the baseline. 
        Defaults to `None`.
    restore_best_weights: bool, optional
        Whether to restore model weights from the epoch with the best value of the monitored quantity. 
        If `False`, the model weights obtained at the last step of training are used. 
        An epoch will be restored regardless of the performance relative to the `baseline`. 
        If no epoch improves on `baseline`, training will run for `patience` epochs 
        and restore weights from the best epoch in that set. 
        Defaults to `False`.
    start_from_epoch: int, optional
        Number of epochs to wait before starting to monitor improvement. 
        This allows for a warm-up period in which no improvement is expected and thus training will not be stopped.
        Defaults to `0`.

    """
    def __init__(
        self, 
        monitor="val_loss",
        min_delta=0, 
        patience=0, 
        verbose=0, 
        mode="auto", 
        threshold=None,
        baseline=None, 
        restore_best_weights=False, 
        start_from_epoch=0
    ):
        super().__init__(monitor, min_delta, patience, verbose, mode, baseline, restore_best_weights, start_from_epoch)
        self.threshold = threshold
        self.threshold_matched_once = False

        # sanity check
        if self.baseline and self.threshold:
            if self.baseline <= self.threshold:
                raise ValueError(f"Baseline must be greater than threshold, received a baseline of{self.baseline} and a threshold of {self.threshold}.")

    def on_epoch_end(self, epoch, logs=None):
        if self.monitor_op is None:
            # Delay setup until the model's metrics are all built
            self._set_monitor_op()

        current = self.get_monitor_value(logs)
        if current is None or epoch < self.start_from_epoch:
            # If no monitor value exists or still in initial warm-up stage.
            return
        if self.restore_best_weights and self.best_weights is None:
            # If best weights were never set,
            # then the current weights are the best.
            self.best_weights = self.model.get_weights()
            self.best_epoch = epoch
        
        # check if threshold is reached. If not, return. If yes, increase wait by one and set flag to true.
        # this code will only be reached once!
        if self.threshold is not None and not self.threshold_matched_once:
            if current < self.threshold:
                self.threshold_matched_once = True
            else:
                return
    
        if self._is_improvement(current, self.best):
            self.best = current
            self.best_epoch = epoch
            if self.restore_best_weights:
                self.best_weights = self.model.get_weights()
            # Only restart wait if we beat both the baseline and our previous
            # best.
            if self.baseline is None or self._is_improvement(current, self.baseline):
                self.wait = 0
                self.threshold_matched_once = True

            return

        self.wait += 1   

        if self.wait >= self.patience and epoch > 0:
            # Patience has been exceeded: stop training
            self.stopped_epoch = epoch
            self.model.stop_training = True



from functools import partial
from tensorflow.python.framework import ops
from tensorflow.python.ops import random_ops
from tensorflow.python.ops.image_ops_impl import _AssertAtLeast3DImage
import tensorflow as tf


""" RANDOM NOISE WITH SNR DATA AUGMENTATION """
def _get_signal_power(image: tf.Tensor, axis: int | tuple) -> tf.Tensor:
    return tf.reduce_sum(abs(image**2), axis=axis, keepdims=True)


def _get_noise(image: tf.Tensor, snr: int | float, axis: int | tuple) -> tf.Tensor:
    noise_power = _get_signal_power(image=image, axis=axis) / (10**(snr / 10))
    noise = tf.random.normal(tf.shape(image))

    return noise / tf.sqrt(_get_signal_power(noise, axis=axis)) * tf.sqrt(noise_power)


def _random_noise(image: tf.Tensor, random_func: callable, scope_name: str):
    with ops.name_scope(None, scope_name, [image]) as scope:
        image = ops.convert_to_tensor(image, name='image')
        image = _AssertAtLeast3DImage(image)
        shape = image.get_shape()

        if shape.ndims == 3:
            # we have unbatched data (t, x, y) without channels.
            # use the same SNR for all data
            axis = (1, 2)
            snr = random_func(shape=[1, ])
        
        elif (shape.ndims == 4) or (shape.ndims == 5):
            # we have batched data (b, t, x, y) with or without channels.
            # get an individual SNR for each sample
            axis=(2, 3)
            snr = random_func(shape=[shape[0], ])
        else:
            raise ValueError(f"Image shape must be either 3 (unbatched) or 4 (batched) dimensions.")
        
        noise = _get_noise(image=image, snr=snr, axis=axis)
        return image + noise


def random_noise(image, snr, seed=None):
    """
    Apply additive white Gaussian noise (AWGN) with SNR randomly sampled from `snr`.
    Can be savely used within `tf.data.Dataset.map()`.

    Parameters
    ----------
    image : tf.Tensor
        image or 2-D tensor to rotate
    snr : int | tuple | list
        (min_snr, max_snr) SNR value in dB for random noise. A single value results in `snr=(1, snr)`.
    seed : int
        Random seed, defaults to `None`.

    Returns
    -------
    image : tf.Tensor
        `image` with AWGN with SNR randomly sampled from (`min_snr`, `max_snr`).

    Raises
    ------
    ValueError
        If more than 2 values for `snr` are provided.
    ValueError
        If unsupported type for `snr` is provided.

    """

    if snr is None:
        return image
    
    if isinstance(snr, (int, float)):
        # we do not want a minimum value of SNR=0, since it will be ONLY noise!
        # best is to provide 2 values...
        minval = min(1, snr)
        maxval = max(1, snr)
    elif isinstance(snr, (list, tuple)):
        if len(snr) > 2:
            raise ValueError(f"Too many values for SNR, received snr={snr}, expected 2 values.")
        if len(snr) == 2:
            minval = min(snr)
            maxval = max(snr)
        else:
            minval = min(1, *snr)
            maxval = max(1, *snr)
    else:
        raise ValueError(f"Unsupported type for snr {type(snr)}.")
    
    random_func = partial(random_ops.random_uniform, minval=minval, maxval=maxval, seed=seed)
    return _random_noise(image=image, random_func=random_func, scope_name="random_noise")


from keras import layers
from importlib import import_module
from functools import partial
from keras import ops
from keras.src.layers.input_spec import InputSpec
from keras.src.backend import standardize_data_format
from keras.src.utils.argument_validation import standardize_tuple


class BaseFFTUpSampling(layers.Layer):
    """
    UpSampling in Forier space

    The input array is transformed via discrete Fourier transform for real-valued signals,
    padded with zeros, and
    transformed to physical space using inverse discrete Fourier transform for real-valued signals.

    Parameters
    ----------
    rank : int
        Rank of the data.
        Determines rank of Fourier transform.
    target_size : int | tuple
        Determines the target size of the output signal.
        Can be smaller than initial input size to perform downsampling in Fourier space.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the data.
        If `None`, this is usually `"channels_last"`, check Keras API!
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, it is derived from the class name.
        Defaults to `None`.
    **kwargs
    
    """

    def __init__(
        self, 
        rank,
        target_size,
        data_format=None,
        name=None, 
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        
        self.target_size = standardize_tuple(target_size, n=self.rank, allow_zero=True, name="size")
        self.data_format = standardize_data_format(data_format)

        # import fft 
        self.rfft_fn = getattr(import_module(name="keras_fft", package=__package__), "rfft" if self.rank == 1 else f"rfft{self.rank}")
        self.irfft_fn = getattr(import_module(name="keras_fft", package=__package__), f"irfft" if self.rank == 1 else f"irfft{self.rank}")
        
    def build(self, input_shape):
        if self.built:
            return

        if self.data_format == "channels_last":
            channel_axis = -1
            input_channel = input_shape[-1]

            # get data axes
            axes = list(range(len(input_shape)))

            # if data format is `"channels_last"`, we have to transpose in order to apply the rfft and irfft along the last axes
            transpose_axes = axes.copy()
            inverse_transpose_axes = axes.copy()

            transpose_axes.insert(1, transpose_axes.pop())
            inverse_transpose_axes.append(inverse_transpose_axes.pop(1))

            self.transpose = partial(ops.transpose, axes=transpose_axes)
            self.inverse_transpose = partial(ops.transpose, axes=inverse_transpose_axes)

        else:
            # NOTE this does not matter too much since currently the layer is restricted to use `"channels_last"` data format!
            channel_axis = 1
            input_channel = input_shape[1]

            # if data format is already `"channels_first"`, we do not have to transpose in order to apply the rfft and irfft
            self.transpose = lambda x: x
            self.inverse_transpose = lambda x: x

        # define input spec
        self.input_spec = InputSpec(
            min_ndim=self.rank + 2, axes={channel_axis: input_channel}
        )

        # define pad with. In Fourier space, we will always have `data_format="channels_first"`!
        # size is a tuple of positive integers that act as a multiplier for the size in each dimension
        # we use RFFT -> shape in Fourier space is 
        # target_shape_fft = target_shape_ph // 2 + 1
        # target_shape_ph = (target_shape_fft + 1) * 2
        # target_shape_fft = initial_shape_fft + padding
        # initial_shape_fft = input_shape or input_shape // 2 + 1 (along last axis when using RFFT)
        # padding   = target_shape_fft - initial_shape_fft
        #           = target_shape_ph - initial_shape_ph                        if axis != -1
        #           = (target_shape_ph // 2 + 1) - (initial_shape_ph // 2 + 1)  else

        # get initial shape along feature axes
        feature_axes = list(range(len(input_shape)))
        feature_axes.pop(channel_axis)  # remove channel axis
        feature_axes.pop(0)  # remove batch dimension
        feature_dims = tuple([input_shape[fa] for fa in feature_axes])

        # Parcivals Theorem: constant energy in frequency spectrum. We have to scale output after padding!
        self.scaling_factor = ops.prod(self.target_size) / ops.prod(feature_dims)

        self.pad_width = (
            (0, 0),
            (0, 0), 
            *[(0, (size // 2 + 1) - (shape // 2 + 1)) if (i == (len(self.target_size) - 1)) else ((size - shape) // 2, (size - shape) // 2) for i, (size, shape) in enumerate(zip(self.target_size, input_shape[(1 if self.data_format == "channels_last" else 2):]))]
        )

        self.built = True

    def rfft(self, x):
        """
        Performs fast Fourier transform on the real-valued `inputs`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Real-valued input tensor.
        
        Returns
        -------
        (y_real, y_imag) : (KerasTensor, KerasTensor)
            Real- and imaginary part of fast Fourier tranform applied to `inputs`.

        Notes
        -----
        `inputs` must be of shape `(batch, channels, *features)`.

        """
        
        x = self.transpose(x)
        x_real, x_imag = self.rfft_fn(x)

        return x_real, x_imag
    
    def irfft(self, inputs):
        """
        Performs inverse fast Fourier transform on the `inputs`.
        
        Parameters
        ----------
        inputs : tuple
            Tuple of `KerasTensor` (real and imaginary part).
        
        Returns
        -------
        outputs : KerasTensor
            Real-valued output of inverse fast Fourier transform applied to `inputs`.

        Notes
        -----
        `inputs` must be of shape `((batch, channels, *features), (batch, channels, *features))`.

        """

        x_real, x_imag = inputs
        y = self.irfft_fn((x_real, x_imag))
        
        return self.inverse_transpose(y)

    # NOTE: this shouldn't be necessary since there are no trainable weights! MAYBE to get FFT backprop correct in TF?
    # def call(self, inputs):
    #     if backend() == "tensorflow":
    #         @ops.custom_gradient
    #         def forward(inputs):
    #             x_real, x_imag = self.rfft(inputs)
                
    #             x_real_padded = ops.pad(x_real, pad_width=self.pad_width)
    #             x_imag_padded = ops.pad(x_imag, pad_width=self.pad_width)

    #             y = self.irfft(x_real_padded, x_imag_padded)

    #             def backprop(dy, variables=None):
    #                 dy_real, dy_imag = self.rfft(dy)

    #                 dy_real_truncated = dy_real[self.truncation_slice]
    #                 dy_imag_truncated = dy_imag[self.truncation_slice]

    #                 dx = self.irfft(dy_real_truncated, dy_imag_truncated)

    #                 return dx, []
                
    #             return y, backprop
            
    #         return forward(inputs)
        
    #     if backend() == "jax":
    #         x_real, x_imag = self.rfft(inputs)
                
    #         x_real_padded = ops.pad(x_real, pad_width=self.pad_width)
    #         x_imag_padded = ops.pad(x_imag, pad_width=self.pad_width)

    #         y = self.irfft(x_real_padded, x_imag_padded)

    #         return y
    
    #     raise NotImplementedError(f"The call method is only implemented for keras backends `'tensorflow'` and `'jax'`")

    def call(self, inputs):
        x_real, x_imag = self.rfft(inputs)
                
        x_real_padded = ops.pad(x_real, pad_width=self.pad_width)
        x_imag_padded = ops.pad(x_imag, pad_width=self.pad_width)

        y = self.irfft((x_real_padded, x_imag_padded))

        # scale the signal
        y *= ops.cast(self.scaling_factor, dtype=y.dtype)

        return y

    def compute_output_shape(self, input_shape):
        input_shape = list(input_shape)

        feature_axes = list(range(len(input_shape)))
        feature_axes.pop(0)  # batch
        feature_axes.pop(-1 if self.data_format == "channels_last" else 0)  # channel dimension

        for fa, size in zip(feature_axes, self.target_size):
            input_shape[fa] = size

        return tuple(input_shape)
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "target_size": self.target_size,
            "data_format": self.data_format
        })

        return config



from keras import saving
from keras import initializers
from .base_fno import BaseFNO


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO1D")
class FNO1D(BaseFNO):
    """
    FNO1D layer, cf. https://arxiv.org/abs/2010.08895

    This layer has two paths:

    The first path applies a spectral convolution to the input, which
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    The second path applies a 1x1 bypass convolution

    Both path are merged with `merge_layer` and passed through an activation function.    

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation function.
        Defaults to `"gelu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FNO1D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format=None, 
        use_bias=True, 
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)), 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=1, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO2D")
class FNO2D(BaseFNO):
    """
    FNO2D layer, cf. https://arxiv.org/abs/2010.08895

    This layer has two paths:

    The first path applies a spectral convolution to the input, which
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    The second path applies a 1x1 bypass convolution

    Both path are merged with `merge_layer` and passed through an activation function.    

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation function.
        Defaults to `"gelu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FNO2D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format=None, 
        use_bias=True, 
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)), 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=2, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="FNO3D")
class FNO3D(BaseFNO):
    """
    FNO3D layer, cf. https://arxiv.org/abs/2010.08895

    This layer has two paths:

    The first path applies a spectral convolution to the input, which
    (1) transforms the layer into Fourier space via discrete Fourier transform for real-valued data (rfft),
    (2) truncates the Fourier modes to the `modes` lowest modes,
    (3) applies the weights (separated into real- and imaginary weights),
    (4) pads the truncated signal in Fourier space back to its initial shape,
    (5) applies inverse discrete Fourier transform for real-valued data (irfft), and
    (6) applies the bias in physical space (if `use_bias=True`)

    The second path applies a 1x1 bypass convolution

    Both path are merged with `merge_layer` and passed through an activation function.    

    Parameters
    ----------
    filters : int
        Number of filters.
    modes : int
        Number of modes after truncation in Fourier space.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation function.
        Defaults to `"gelu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Merge operation in FNO layers to combine the result from the spectral convolution with the result from the bypass convolution.
        Defaults to `"add"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Format of the input data.
        Defaults to `None`.
    use_bias : bool, optional
        If `True`, a bias term is added to the physical space after the spectral convolution.
        Defaults to `True`.
    kernel_initializer : str | keras.initializer.Initializer | tuple, optional
        Initializer for real- and imaginary weights.
        By default, the real weights are initialized using `"glorot_normal"`,
        and the imaginary weights are initialized using `"random"` with a low standard deviation,
        which is effectively white Gaussian noise.
        Defaults to `("glorot_normal", initializers.RandomNormal(stddev=1e-3))`.
    bias_initializer : str | keras.initializer.Initializer, optional
        Initializer for the bias.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"FNO3D"`.
        Defaults to `None`.

    Notes
    -----
    For implementataion simplicity, the Fourier operations are always performed in `"channels_first"` data format.
    The layer therefore applies a transpose operation if `data_format="channels_last"`.
    
    """

    def __init__(
        self, 
        filters, 
        modes, 
        activation="gelu", 
        merge_layer="add", 
        data_format=None, 
        use_bias=True, 
        kernel_initializer=("glorot_normal", initializers.RandomNormal(stddev=1e-3)), 
        bias_initializer="zeros", 
        kernel_constraint=None, 
        bias_constraint=None, 
        kernel_regularizer=None, 
        bias_regularizer=None, 
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=3, 
            filters=filters, 
            modes=modes, 
            activation=activation, 
            merge_layer=merge_layer,  
            data_format=data_format, 
            use_bias=use_bias, 
            kernel_initializer=kernel_initializer, 
            bias_initializer=bias_initializer, 
            kernel_constraint=kernel_constraint, 
            bias_constraint=bias_constraint, 
            kernel_regularizer=kernel_regularizer, 
            bias_regularizer=bias_regularizer, 
            name=name, 
            **kwargs
        )
        


from. autoencoder import FCN1D, FCN2D, FCN3D
from .autoencoder import Unet1D, Unet2D, Unet3D, SmoothUnet1D, SmoothUnet2D, SmoothUnet3D

from .neural_operator import NeuralOperator1D, NeuralOperator2D

from .tsmixer import TSMixer


from keras import saving
from keras import layers


@saving.register_keras_serializable(package="Kerex.Blocks", name="TSMixerBlock")
class TSMixerBlock(layers.Layer):
    """
    TSMixer block, cf. [Chen et al.](https://doi.org/10.48550/arXiv.2303.06053)

    Parameters
    ----------
    num_hidden : int, optional
        Number of units in dense layer.
        If `None`, the feature space is expanded by factor of 4.
        Defaults to `None`.
    norm : str, optional {`"LN"`, `"BN"`}
        Normalization type, can be layer normalization (`"LN"`) or batch normalization (`"BN"`).
        Defaults to `"LN"`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for dense layers.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    dropout_rate : float, optional
        Dropout rate.
        Defaults to 0.1.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"TSMixerBlock"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    Raises
    ------
    ValueError
        If `norm` other than `"LN"` or `"BN"` is provided.

    Notes
    -----
    Adapted for Keras3 from https://github.com/google-research/google-research/tree/master/tsmixer

    """

    def __init__(
        self,
        num_hidden=None,
        norm="LN",
        activation="relu",
        dropout_rate=0.1,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.num_hidden = num_hidden
        self.norm = norm
        self.activation = activation
        self.dropout_rate = dropout_rate

        # define layers
        if self.norm == "LN":
            self.temporal_norm = layers.LayerNormalization(axis=(-2, -1), name="temporal_norm")
            self.feature_norm = layers.LayerNormalization(axis=(-2, -1), name="feature_norm")
        elif self.norm == "BN":
            self.temporal_norm = layers.BatchNormalization(axis=-2, name="temporal_norm")
            self.feature_norm = layers.BatchNormalization(axis=-1, name="feature_norm")
        else:
            raise ValueError(f"Unknown `norm` '{self.norm}'. Valid values for `norm` are `'LN'` (layer norm) and `'BN'` (batch norm).")

        self.transpose_1 = layers.Permute(dims=(2, 1), name="transpose_1")
        self.transpose_2 = layers.Permute(dims=(2, 1), name="transpose_2")

        # temporal linear layers
        self.temporal_dense = None  # depends on input shape! move to build method
        self.temporal_dropout = layers.Dropout(rate=self.dropout_rate)

        # feature linear layers
        self.feature_dense_1 = None
        self.feature_dropout_1 = layers.Dropout(rate=self.dropout_rate)
        self.feature_dense_2 = None  # depends on input shape! move to build method
        self.feature_dropout_2 = layers.Dropout(rate=self.dropout_rate)

    def build(self, input_shape):
        if self.built:
            return

        # build layers sequentually by building the layers and updating the input shape
        self.temporal_norm.build(input_shape=input_shape)
        forward_shape = self.temporal_norm.compute_output_shape(input_shape=input_shape)

        self.transpose_1.build(input_shape=forward_shape)
        forward_shape = self.transpose_1.compute_output_shape(input_shape=forward_shape)

        # temporal_dense is not defined yet
        self.temporal_dense = layers.Dense(units=forward_shape[-1], activation=self.activation, name="temporal_dense")
        self.temporal_dense.build(input_shape=forward_shape)
        forward_shape = self.temporal_dense.compute_output_shape(input_shape=forward_shape)

        self.transpose_2.build(input_shape=forward_shape)
        forward_shape = self.transpose_2.compute_output_shape(input_shape=forward_shape)

        # dropout does not modify `forward-shape`
        self.temporal_dropout.build(input_shape=forward_shape)

        # addition does not alter the `forward_shape` either!
        self.feature_norm.build(input_shape=forward_shape)
        forward_shape = self.feature_norm.compute_output_shape(input_shape=forward_shape)

        # feature_dense_1 is not defined yet
        self.feature_dense_1 = layers.Dense(units=self.num_hidden or 4 * forward_shape[-1], activation=self.activation, name="feature_dense_1")
        self.feature_dense_1.build(input_shape=forward_shape)
        forward_shape = self.feature_dense_1.compute_output_shape(input_shape=forward_shape)

        self.feature_dropout_1.build(input_shape=forward_shape)

        # feature_dense_2 is not defined yet
        self.feature_dense_2 = layers.Dense(units=input_shape[-1], name="feature_dense_2")
        self.feature_dense_2.build(input_shape=forward_shape)
        forward_shape = self.feature_dense_2.compute_output_shape(input_shape=forward_shape)

        self.feature_dropout_2.build(input_shape=forward_shape)

        self.built = True

    def call(self, inputs):
        # temporal linear
        x = self.temporal_norm(inputs)
        x = self.transpose_1(x)
        x = self.temporal_dense(x)
        x = self.transpose_2(x)
        x = self.temporal_dropout(x)

        res = x + inputs

        # feature linear
        x = self.feature_norm(res)
        x = self.feature_dense_1(x)
        x = self.feature_dropout_1(x)
        x = self.feature_dense_2(x)
        x = self.feature_dropout_2(x)

        return x + res
    
    def compute_output_shape(self, input_shape):
        forward_shape = self.temporal_norm.compute_output_shape(input_shape=input_shape)
        forward_shape = self.transpose_1.compute_output_shape(input_shape=forward_shape)
        forward_shape = self.temporal_dense.compute_output_shape(input_shape=forward_shape)
        forward_shape = self.transpose_2.compute_output_shape(input_shape=forward_shape)
        forward_shape = self.feature_norm.compute_output_shape(input_shape=forward_shape)
        forward_shape = self.feature_dense_1.compute_output_shape(input_shape=forward_shape)
        forward_shape = self.feature_dense_2.compute_output_shape(input_shape=forward_shape)

        return forward_shape
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "num_hidden": self.num_hidden,
            "norm": self.norm,
            "activation": saving.serialize_keras_object(self.activation),
            "dropout_rate": self.dropout_rate
        })

        return config
    
    @classmethod
    def from_config(cls, config):
        activation_cfg = config.pop("activation")
        config.update({"activation": saving.deserialize_keras_object(activation_cfg)})

        return cls(**config)
    


from .fno import FNO1D, FNO2D, FNO3D

from .merge import AttentionGate

from .pooling import ChannelAttentionModule1D, ChannelAttentionModule2D, ChannelAttentionModule3D

from .reshape import SmoothUpSampling1D, SmoothUpSampling2D, SmoothUpSampling3D, FFTUpSampling1D, FFTUpSampling2D, FFTUpSampling3D

from .wrapper import SpatialAttention, ChannelAttention, CBAM, FiLM, Residual, TemporalSlice, WeightNormalization


from keras import layers
from keras import Sequential
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import saving
from keras.src.layers.merging.base_merge import Merge
from ...ops.helper import _IterableVars
from ...ops import get_layer
from importlib import import_module
from keras.src.backend import standardize_data_format


class BaseDecoder(layers.Layer, _IterableVars):
    """
    Base class of convolutional decoder block

    Use to subclass 1-D, 2-D, and 3-D Decoder

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseEncoder`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    upsampling_filters : int, optional
        Number of filters for the upsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from the second input.
        Defaults to `"concatenate"`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"BaseDecoder"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    Raises
    ------
    TypeError
        If `merge_layer` is not valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/
        
    """

    def __init__(
            self,
            rank,
            filters,
            kernel_size=5,
            strides=1,
            padding="same",
            data_format="channels_last",
            dilation_rate=1,
            groups=1,
            upsampling_filters=None,
            activation="relu",
            use_bias=True,
            merge_layer="concatenate",
            kernel_initializer="he_normal",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            name=None, 
            **kwargs):
        super().__init__(name=name, **kwargs)

        # set general class variables
        self.rank = rank
        self.padding = padding
        self.data_format = standardize_data_format(data_format)
        self.activation = activation
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint

        # set class variables for the Sequential model
        self.set_vars(filters=filters, kernel_size=kernel_size, strides=strides, dilation_rate=dilation_rate, groups=groups)        

        # set class variables that are relevant for upsampling operation
        self.upsampling_filters = upsampling_filters or self.filters[0]

        # define layers
        self.forward_conv = Sequential([
            getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=f,
                kernel_size=k,
                strides=s,
                padding=self.padding,
                data_format=self.data_format,
                dilation_rate=d,
                groups=g,
                activation=self.activation,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                activity_regularizer=self.activity_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint
            ) for f, k, s, d, g in zip(self.filters, self.kernel_size, self.strides, self.dilation_rate, self.groups)
        ])

        self.upsampling = getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}DTranspose")(
            filters=self.upsampling_filters,
            kernel_size=2,
            strides=2,
            padding="same",
            data_format=self.data_format,
            dilation_rate=1,
            activation=None,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            activity_regularizer=self.activity_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint
        )

        # load merge layer
        try:
            self.merge_layer = get_layer(merge_layer, axis=-1 if self.data_format == "channels_last" else 1)
        except TypeError:  # layer does not supply axis argument
            self.merge_layer = get_layer(merge_layer)

        if not issubclass(type(self.merge_layer), Merge):
            raise TypeError(
                f"Merge-layer {self.merge_layer} supplied to Wrapper isn't "
                "a supported merge-layer."
            )
        
    def call(self, inputs, skip=None):
        """
        Call method of `BaseDecoder`

        The `inputs` are first processed by `self.forward_conv`,
        and then upsampled by factor of 2 along the feature axes by `self.upsampling`.
        If an additional input `skip` is provided, 
        the upsampled Tensor is merged with the `skip` Tensor using `self.merge_layer`.
        
        Parameters
        ----------
        inputs : KerasTensor
            Input tensor.
        skip : KerasTensor, optional
            Additional Tensor that is merged with the forward Tensor
            Defaults to `None`.

        Returns
        -------
        outputs : KerasTensor
            Output tensor.
            
        """

        x_forward = self.upsampling(inputs)

        if skip is not None:
            x_forward = self.merge_layer((x_forward, skip))

        x_forward = self.forward_conv(x_forward)

        return x_forward
    
    def build(self, input_shape, input_shape_skip=None):
        """
        Build method of `BaseDecoder`

        Builds all sub-modules and sets `self.built=True`.
        
        Parameters
        ----------
        input_shape : tuple
            Input shape
        input_shape_skip : tuple, optional
            Input shape of additional information from skip connection.
            Defaults to `None`.

        """

        if self.built:
            return
        
        # cache input shapes
        # self._build_shapes_dict = {"input_shape": input_shape, "input_shape_skip": input_shape_skip}
        self._build_shapes = {}

        super().build(input_shape)

        # build upsampling layer
        # cache build shape
        self._build_shapes.update({self.upsampling.name: {"input_shape": input_shape}})

        self.upsampling.build(input_shape=input_shape)
        input_shape = self.upsampling.compute_output_shape(input_shape=input_shape)

        # build merge layer
        if input_shape_skip is not None:
            # cache build shape
            self._build_shapes.update({self.merge_layer.name: {"input_shape": (input_shape, input_shape_skip)}})

            self.merge_layer.build(input_shape=(input_shape, input_shape_skip))
            input_shape = self.merge_layer.compute_output_shape(input_shape=(input_shape, input_shape_skip))

        # build forward layer and update input_shape
        # cache build shape
        self._build_shapes.update({self.forward_conv.name: {"input_shape": input_shape}})

        self.forward_conv.build(input_shape=input_shape)

        # update built state
        self.built = True
    
    def compute_output_shape(self, input_shape, input_shape_skip=None):
        """
        Compute output shape of `BaseDecoder`

        Parameters
        ----------
        input_shape : tuple
            Input shape
        input_shape_skip : tuple, optional
            Input shape of additional information from skip connection.
            Defaults to `None`.

        Returns
        -------
        output_shape : tuple
            Output shape.
            
        """

        input_shape = self.upsampling.compute_output_shape(input_shape=input_shape)
        if input_shape_skip is not None:
            input_shape = self.merge_layer.compute_output_shape(input_shape=(input_shape, input_shape_skip))

        output_shape = self.forward_conv.compute_output_shape(input_shape=input_shape)
        
        return output_shape

    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.

        """

        config: dict = super().get_config()
        config.update({
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "groups": self.groups,
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "upsampling_filters": self.upsampling_filters,
            "activation": saving.serialize_keras_object(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint)
        })
        
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BaseDecoder
            The `BaseDecoder` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseDecoder
            Instance of `BaseDecoder` from `config`.
        """

        # get configs of keras objects
        merge_layer_cfg = config.pop("merge_layer")
        activation_cfg = config.pop("activation")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")

        config.update({
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "activation": saving.deserialize_keras_object(activation_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg)
        })

        return cls(**config)



# BaseDecoder with smooth upsampling
class BaseSmoothDecoder(BaseDecoder):
    """
    Base class of convolutional decoder block

    Use to subclass 1-D, 2-D, and 3-D Decoder

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseEncoder`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters for the convolutional forward sub-model.
    kernel_size : int | list | tuple, optional
        Kernel size for the convolutional forward sub-model.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for the convolutional forward sub-model.
        Defaults to 1.
    padding : str, optional
        Padding that is applied to maintain deterministic data shapes.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rate for the convolutional forward sub-model.
        Defaults to 1.
    groups : int | list | tuple, optional
        Number of groups rate for the convolutional forward sub-model.
        Defaults to 1.
    upsampling_filters : int, optional
        Number of filters for the upsampling operation.
        If `None`, this parameter is set to the last entry of `filters`.
        Defaults to `None`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Activation for the convolutional forward sub-model.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from the second input.
        Defaults to `"concatenate"`.
    use_bias : bool, optional
        Whether to use bias.
        Defaults to `True`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the layer.
        If `None`, `name` is automatically inherited from the class name `"BaseDecoder"`.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `keras.layers.Layer` super-class.

    Raises
    ------
    TypeError
        If `merge_layer` is not valid Keras merge layer, cf. https://keras.io/api/layers/merging_layers/
        
    """

    def __init__(
            self,
            rank,
            filters,
            kernel_size=5,
            strides=1,
            padding="same",
            data_format="channels_last",
            dilation_rate=1,
            groups=1,
            upsampling_filters=None,
            upsampling_kernel_size=None,
            activation="relu",
            use_bias=True,
            merge_layer="concatenate",
            interpolation="nearest",
            kernel_initializer="he_normal",
            bias_initializer="zeros",
            kernel_regularizer=None,
            bias_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            upsampling_filters=upsampling_filters,
            activation=activation,
            use_bias=use_bias,
            merge_layer=merge_layer,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )

        # overwrite upsampling layer
        self.upsampling_kernel_size = upsampling_kernel_size or self.kernel_size[0]
        self.interpolation = interpolation
        
        self.upsampling = getattr(import_module(name="...layers.reshape", package=__package__), f"SmoothUpSampling{self.rank}D")(
            filters=self.upsampling_filters,
            kernel_size=self.upsampling_kernel_size,
            size=tuple([2] * self.rank),
            strides=1,
            padding="same",
            data_format=self.data_format,
            use_bias=self.use_bias,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            activity_regularizer=self.activity_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
            interpolation=self.interpolation  # only for 2D!
        )

    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.

        """

        config: dict = super().get_config()
        config.update({
            "interpolation": self.interpolation,
            "upsampling_kernel_size": self.upsampling_kernel_size
        })
        
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BaseDecoder
            The `BaseDecoder` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseDecoder
            Instance of `BaseDecoder` from `config`.
        """

        # get configs of keras objects
        merge_layer_cfg = config.pop("merge_layer")
        activation_cfg = config.pop("activation")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")

        config.update({
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "activation": saving.deserialize_keras_object(activation_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg)
        })

        return cls(**config)


from keras import layers
from keras import ops
from keras import saving


@saving.register_keras_serializable(package="Kerex.Layers.Wrapper", name="WeightNormalization")
class WeightNormalization(layers.Wrapper):
    """
    Weight Normalization wrapper for any Keras `layer`.
    
    The weight normalization is supposed to accelerate the convergence of stochastic gradient descent optimization.
    It replaces the original weights `w` with a normalized weight `v` that is multiplied by a bias `g`:
    `w -> g * v`
    The original layer bias `b` remains untouched.

    Parameters
    ----------
    layer : Layer
        Keras `layer` to be wrapped.
    name : str, optional
        Name of the wrapper, that is prepended to the name of the wrapped layer.
        Defaults to `"weight_normalization"`.
    **kwargs : Additional keyword arguments for the `Wrapper` super-class.

    Notes
    -----
    Reference: Talisman et al., Weight Normalization: A Simple Reparameterization to Accelerate Training of Deep Neural Networks
    https://arxiv.org/abs/1602.07868

    """

    def __init__(self, layer, name="weight_normalization", **kwargs): 
        name = '_'.join([layer.name, name])
        super().__init__(layer=layer, name=name, **kwargs)

    def build(self, input_shape):
        """
        Build method of `WeightNormalization`.

        Builds all sub-modules and sets `self.built=True`.

        Parameters
        ----------
        input_shape : tuple
            Shape of the input data.

        """

        super().build(input_shape=input_shape)
        self.v = self.add_weight(shape=self.layer.kernel.shape, initializer=self.layer.kernel_initializer, trainable=True)
        self.g = self.add_weight(shape=self.layer.bias.shape, initializer='ones', trainable=True)

    def call(self, inputs):
        """
        Call method of `WeightNormalization`

        The weight normalization is supposed to accelerate the convergence of stochastic gradient descent optimization.
        It replaces the original weights `w` with a normalized weight `v` that is multiplied by a bias `g`:
        `w -> g * v`
        The original layer bias `b` remains untouched.

        Parameters
        ----------
        inputs : KerasTensor
            Input tensor for the wrapped Keras `layer`.

        Returns
        -------
        outputs : KerasTensor
            Output tensor of the wrapped Keras `layer``.

        """
        norm_v = ops.normalize(self.v, axis=0)
        self.layer.set_weights([self.g * norm_v, self.layer.bias])
        return self.layer(inputs)
    
    def compute_output_shape(self, input_shape: tuple) -> tuple:
        """
        Compute output shape

        Parameters
        ----------
        input_shape : tuple
            Shape of the input.

        Returns
        -------
        output_shape : tuple
            Shape of the output of `WeightNormalization`.

        """

        return self.layer.compute_output_shape(input_shape)
    


from .base_smooth_upsampling import BaseSmoothUpSampling
from keras import layers
from keras import saving


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="SmoothUpSampling1D")
class SmoothUpSampling1D(BaseSmoothUpSampling):
    def __init__(
        self, 
        filters, 
        kernel_size, 
        size=2,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None, 
        **kwargs
    ):
        kwargs.pop("interpolation", None)
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name, 
            **kwargs
        )

        self.upsampling = layers.UpSampling1D(size=size)


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="SmoothUpSampling2D")
class SmoothUpSampling2D(BaseSmoothUpSampling):
    def __init__(
        self, 
        filters, 
        kernel_size, 
        size=(2, 2),
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        interpolation="nearest",
        name=None, 
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name, 
            **kwargs
        )
        
        self.upsampling = layers.UpSampling2D(
            size=size, 
            interpolation=interpolation,
            data_format=data_format, 
        )


@saving.register_keras_serializable(package="Kerex.Layers.Reshape", name="SmoothUpSampling3D")
class SmoothUpSampling3D(BaseSmoothUpSampling):
    def __init__(
        self, 
        filters, 
        kernel_size, 
        size=(2, 2, 2),
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None, 
        **kwargs
    ):
        kwargs.pop("interpolation", None)
        super().__init__(
            rank=3,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name, 
            **kwargs
        )

        self.upsampling = layers.UpSampling3D(
            size=size, 
            data_format=data_format, 
        )



from keras import saving
from .base_cam import BaseChannelAttentionModule


@saving.register_keras_serializable(package="Kerex.Layers.Pooling", name="ChannelAttentionModule1D")
class ChannelAttentionModule1D(BaseChannelAttentionModule):
    """
    Convolutional Channel Attention Module (CAM)

    Parameters
    ----------
    reduction_ratio : int, optional
        Size of the bottleneck of the shared MLP in relation to the input size.
        Defaults to 8.
    activation : str, optional
        Activation function for the shared MLP.
        Defaults to `"sigmoid"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format of the convolutions.
        Can be either `"channels_first"` or `"channels_last"`.
        Defaults to `"channels_last"`.
    keepdims : bool, optional
        If `True`, the dimensions are preserved (as 1) throughout the pooling process.
        Select `False` to get a global pooling layer that automatically weighs the importance of channels.
        Defaults to `True`.
    name : str, optional
        Name of the layer.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `Layer` super-class.

    Notes
    -----
    This is not a wrapper but a CAM layer.
    May be used on it's own, but here it serves as a building block for the 
    [Convolutional Block Attention Module (CBAM)](https://arxiv.org/abs/1807.06521).
    
    More information on the CAM can be found here: https://www.digitalocean.com/community/tutorials/attention-mechanisms-in-computer-vision-cbam#channel-attention-module-cam
    
    """

    def __init__(
            self, 
            reduction_ratio=8, 
            activation="sigmoid", 
            data_format="channels_last", 
            keepdims=True, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=1,
            reduction_ratio=reduction_ratio,
            activation=activation,
            data_format=data_format,
            keepdims=keepdims,
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.Pooling", name="ChannelAttentionModule2D")
class ChannelAttentionModule2D(BaseChannelAttentionModule):
    """
    Convolutional Channel Attention Module (CAM)

    Parameters
    ----------
    reduction_ratio : int, optional
        Size of the bottleneck of the shared MLP in relation to the input size.
        Defaults to 8.
    activation : str, optional
        Activation function for the shared MLP.
        Defaults to `"sigmoid"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format of the convolutions.
        Can be either `"channels_first"` or `"channels_last"`.
        Defaults to `"channels_last"`.
    keepdims : bool, optional
        If `True`, the dimensions are preserved (as 1) throughout the pooling process.
        Select `False` to get a global pooling layer that automatically weighs the importance of channels.
        Defaults to `True`.
    name : str, optional
        Name of the layer.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `Layer` super-class.

    Notes
    -----
    This is not a wrapper but a CAM layer.
    May be used on it's own, but here it serves as a building block for the 
    [Convolutional Block Attention Module (CBAM)](https://arxiv.org/abs/1807.06521).
    
    More information on the CAM can be found here: https://www.digitalocean.com/community/tutorials/attention-mechanisms-in-computer-vision-cbam#channel-attention-module-cam
    
    """

    def __init__(
            self, 
            reduction_ratio=8, 
            activation="sigmoid", 
            data_format="channels_last", 
            keepdims=True, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=2,
            reduction_ratio=reduction_ratio,
            activation=activation,
            data_format=data_format,
            keepdims=keepdims,
            name=name, 
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.Pooling", name="ChannelAttentionModule3D")
class ChannelAttentionModule3D(BaseChannelAttentionModule):
    """
    Convolutional Channel Attention Module (CAM)

    Parameters
    ----------
    reduction_ratio : int, optional
        Size of the bottleneck of the shared MLP in relation to the input size.
        Defaults to 8.
    activation : str, optional
        Activation function for the shared MLP.
        Defaults to `"sigmoid"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format of the convolutions.
        Can be either `"channels_first"` or `"channels_last"`.
        Defaults to `"channels_last"`.
    keepdims : bool, optional
        If `True`, the dimensions are preserved (as 1) throughout the pooling process.
        Select `False` to get a global pooling layer that automatically weighs the importance of channels.
        Defaults to `True`.
    name : str, optional
        Name of the layer.
        Defaults to `None`.
    **kwargs : Additional keyword arguments for the `Layer` super-class.

    Notes
    -----
    This is not a wrapper but a CAM layer.
    May be used on it's own, but here it serves as a building block for the 
    [Convolutional Block Attention Module (CBAM)](https://arxiv.org/abs/1807.06521).
    
    More information on the CAM can be found here: https://www.digitalocean.com/community/tutorials/attention-mechanisms-in-computer-vision-cbam#channel-attention-module-cam
    
    """

    def __init__(
            self, 
            reduction_ratio=8, 
            activation="sigmoid", 
            data_format="channels_last", 
            keepdims=True, 
            name=None, 
            **kwargs
        ):
        super().__init__(
            rank=3,
            reduction_ratio=reduction_ratio,
            activation=activation,
            data_format=data_format,
            keepdims=keepdims,
            name=name, 
            **kwargs
        )



from keras import saving
from .base_attention import BaseAttention


@saving.register_keras_serializable(package="Kerex.Blocks.Attention", name="GlobalSelfAttention")
class GlobalSelfAttention(BaseAttention):
    """
    Global self-attention layer.
    
    This layer calculates the self-attention of the `query`.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    key_dim : int
        Size of each attention head for query and key.
    value_dim : int, optional
        Size of each attention head for value.
        If this is `None`, it is set to `key_dim`.
        Defaults to `None`.
    dropout : float, optional
        Dropout probability. Defaults to 0.
    use_bias : bool, optional
        If this is set, a bias is ued. Defaults to `True`.
    output_shape : tuple, optional
        The expected shape of an output tensor, besides the batch and sequence dims.
        If not specified, projects back to the query feature dim (the query input's last dimension).
        Defaults to `None`.
    attention_axes : int, optional
        Axes over which the attention is applied.
        `None` means attention over all axes, but batch, heads, and features.
        Defaults to `None`.
    kernel_initializer : str, optional
        Initializer for dense layer kernels. Defaults to `"glorot_uniform"`.
    bias_initializer : str, optional
        Initializer for dense layer biases. Defaults to `"zeros"`.
    kernel_regularizer : str, optional
        Regularizer for dense layer kernels. Defaults to `None`.
    bias_regularizer : str, optional
        Regularizer for dense layer biases. Defaults to `None`.
    activity_regularizer : str, optional
        Regularizer for dense layer activity. Defaults to `None`.
    kernel_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    bias_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    seed : int, optional
        Optional integer to seed the dropout layer. Defaults to `None`

    Notes
    -----
    With flash attention (Keras>3.6.0), the dropout probability has to be 0.

    """

    def call(self, x, return_attention_scores=False):
        """
        Call method of GlobalSelfAttention layer

        Parameters
        ----------
        x : KerasTensor
            Tensor for which the self-attention is calculated.
        return_attention_scores : bool, optional
            If `True`, the attention scores are returned. 
            Defaults to `False`.

        Returns
        -------
        y | (y, attention_scores) : KerasTensor | (KerasTensor, KerasTensor)
            The output of the self-attention.
            If `return_attention_scores==True`, the layer additionally returns the attention scores.

        """

        attn_output = self.mha(
            query=x,
            value=x,
            key=x,
            return_attention_scores=return_attention_scores
        )
        if return_attention_scores:
            attn_output, attn_scores = attn_output

        x = self.add([x, attn_output])
        x = self.layernorm(x)

        if return_attention_scores:
            return x, attn_scores
        return x


@saving.register_keras_serializable(package="Kerex.Blocks.Attention", name="CausalSelfAttention")
class CausalSelfAttention(BaseAttention):
    """
    Causal self-attention layer.
    
    This layer calculates the self-attention of the `query`.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    key_dim : int
        Size of each attention head for query and key.
    value_dim : int, optional
        Size of each attention head for value.
        If this is `None`, it is set to `key_dim`.
        Defaults to `None`.
    dropout : float, optional
        Dropout probability. Defaults to 0.
    use_bias : bool, optional
        If this is set, a bias is ued. Defaults to `True`.
    output_shape : tuple, optional
        The expected shape of an output tensor, besides the batch and sequence dims.
        If not specified, projects back to the query feature dim (the query input's last dimension).
        Defaults to `None`.
    attention_axes : int, optional
        Axes over which the attention is applied.
        `None` means attention over all axes, but batch, heads, and features.
        Defaults to `None`.
    kernel_initializer : str, optional
        Initializer for dense layer kernels. Defaults to `"glorot_uniform"`.
    bias_initializer : str, optional
        Initializer for dense layer biases. Defaults to `"zeros"`.
    kernel_regularizer : str, optional
        Regularizer for dense layer kernels. Defaults to `None`.
    bias_regularizer : str, optional
        Regularizer for dense layer biases. Defaults to `None`.
    activity_regularizer : str, optional
        Regularizer for dense layer activity. Defaults to `None`.
    kernel_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    bias_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    seed : int, optional
        Optional integer to seed the dropout layer. Defaults to `None`

    Notes
    -----
    With flash attention (Keras>3.6.0), the dropout probability has to be 0.

    """

    def call(self, x, prefix=0, prediction_step=0, return_attention_scores=False):
        """
        Call method of GlobalSelfAttention layer

        Parameters
        ----------
        x : KerasTensor
            Tensor for which the self-attention is calculated.
        prefix : int, optional
            Modifies the attention mask if `prefix>0`.
            Defaults to 0.
        prediction_step : int, optional
            Modifies the attention mask if `prediction_step>0`.
            Defaults to 0.
        return_attention_scores : bool, optional
            If `True`, the attention scores are returned. 
            Defaults to `False`.

        Returns
        -------
        y | (y, attention_scores) : KerasTensor | (KerasTensor, KerasTensor)
            The output of the self-attention.
            If `return_attention_scores==True`, the layer additionally returns the attention scores.

        """
        
        attn_output = self.mha(
            query=x,
            value=x,
            key=x,
            use_causal_mask=True,
            prefix=prefix,
            prediction_step=prediction_step,
            return_attention_scores=return_attention_scores
        )

        if return_attention_scores:
            attn_output, attn_scores = attn_output

        x = self.add([x, attn_output])
        x = self.layernorm(x)

        if return_attention_scores:
            return x, attn_scores
        return x


@saving.register_keras_serializable(package="Kerex.Blocks.Attention", name="CrossAttention")
class CrossAttention(BaseAttention):
    """
    Cross-attention layer.
    
    This layer calculates the cross-attention of the `query` and the `key`.
    The `key` represents the context that is merged with an input.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    key_dim : int
        Size of each attention head for query and key.
    value_dim : int, optional
        Size of each attention head for value.
        If this is `None`, it is set to `key_dim`.
        Defaults to `None`.
    dropout : float, optional
        Dropout probability. Defaults to 0.
    use_bias : bool, optional
        If this is set, a bias is ued. Defaults to `True`.
    output_shape : tuple, optional
        The expected shape of an output tensor, besides the batch and sequence dims.
        If not specified, projects back to the query feature dim (the query input's last dimension).
        Defaults to `None`.
    attention_axes : int, optional
        Axes over which the attention is applied.
        `None` means attention over all axes, but batch, heads, and features.
        Defaults to `None`.
    kernel_initializer : str, optional
        Initializer for dense layer kernels. Defaults to `"glorot_uniform"`.
    bias_initializer : str, optional
        Initializer for dense layer biases. Defaults to `"zeros"`.
    kernel_regularizer : str, optional
        Regularizer for dense layer kernels. Defaults to `None`.
    bias_regularizer : str, optional
        Regularizer for dense layer biases. Defaults to `None`.
    activity_regularizer : str, optional
        Regularizer for dense layer activity. Defaults to `None`.
    kernel_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    bias_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    seed : int, optional
        Optional integer to seed the dropout layer. Defaults to `None`

    Notes
    -----
    With flash attention (Keras>3.6.0), the dropout probability has to be 0.

    """

    def call(self, x, context, return_attention_scores=False):
        """
        Call method of GlobalSelfAttention layer

        Parameters
        ----------
        x : KerasTensor
            Input tensor (query).
        context : KerasTensor
            Context for `x` to calculate the attention.
        return_attention_scores : bool, optional
            If `True`, the attention scores are returned. 
            Defaults to `False`.

        Returns
        -------
        y | (y, attention_scores) : KerasTensor | (KerasTensor, KerasTensor)
            The output of the self-attention.
            If `return_attention_scores==True`, the layer additionally returns the attention scores.

        """
        attn_output = self.mha(
            query=x,
            key=context,
            value=context,
            return_attention_scores=return_attention_scores
        )

        if return_attention_scores:
            attn_output, attn_scores = attn_output

        x = self.add([x, attn_output])
        x = self.layernorm(x)

        if return_attention_scores:
            return x, attn_scores
        return x



from keras import optimizers
from keras.src import backend
from keras.src import ops
from keras import saving


@saving.register_keras_serializable(package="Kerex.Optimizers", name="AdamW")
class AdamW(optimizers.Adam):
    """
    Optimizer that implements the AdamW algorithm.
    
    "AdamW is often superior to Adam with L2 regularization because it decouples
    the weight decay from the gradient-based updates, leading to more effective 
    regularization, better generalization, and improved convergence. 
    This makes AdamW a more robust and reliable optimizer for a wide range of 
    deep learning tasks."
    (https://www.geeksforgeeks.org/why-is-adamw-often-superior-to-adam-with-l2-regularization-in-practice/)

    AdamW optimization is a stochastic gradient descent method that is based on
    adaptive estimation of first-order and second-order moments with an added
    method to decay weights per the techniques discussed in the paper,
    'Decoupled Weight Decay Regularization' by
    [Loshchilov, Hutter et al., 2019](https://arxiv.org/abs/1711.05101).

    According to
    [Kingma et al., 2014](http://arxiv.org/abs/1412.6980),
    the underying Adam method is "*computationally
    efficient, has little memory requirement, invariant to diagonal rescaling of
    gradients, and is well suited for problems that are large in terms of
    data/parameters*".

    Parameters
    ----------
    learning_rate : float | keras.optimizers.schedules.LearningRateSchedule | callable
        The learning rate or a learning rate-scheduler. Defaults to `0.001`.
    beta_1 : float | KerasTensor | callable, optional
        The exponential decay rate for the 1st moment estimates.
        Defaults to `0.9`.
    beta_2 : float | KerasTensor | callable, optional
        The exponential decay rate for the 2nd moment estimates.
        Defaults to `0.999`.
    epsilon : float, optional
        A small constant for numerical stability. This epsilon is
        "epsilon hat" in the Kingma and Ba paper (in the formula just
        before Section 2.1), not the epsilon in Algorithm 1 of the paper.
        Defaults to 1e-7.
    amsgrad : bool, optional
        Whether to apply AMSGrad variant of this algorithm
        from the paper "On the Convergence of Adam and beyond".
        Defaults to `False`.    

    Notes
    -----
    In constrast to the default Keras implementation of AdamW, the weight
    decay is decoupled from the gradient update process!

    References:
    - [Loshchilov et al., 2019](https://arxiv.org/abs/1711.05101)
    - [Kingma et al., 2014](http://arxiv.org/abs/1412.6980) for `adam`
    - [Reddi et al., 2018](https://openreview.net/pdf?id=ryQu7f-RZ) for `amsgrad`.

    """

    def __init__(
        self,
        learning_rate=0.001,
        weight_decay=0.0001,
        beta_1=0.9,
        beta_2=0.999,
        epsilon=1e-7,
        amsgrad=False,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
        loss_scale_factor=None,
        gradient_accumulation_steps=None,
        name="adamw",
        **kwargs,
    ):
        super().__init__(
            learning_rate=learning_rate,
            beta_1=beta_1,
            beta_2=beta_2,
            epsilon=epsilon,
            amsgrad=amsgrad,
            name=name,
            weight_decay=weight_decay,
            clipnorm=clipnorm,
            clipvalue=clipvalue,
            global_clipnorm=global_clipnorm,
            use_ema=use_ema,
            ema_momentum=ema_momentum,
            ema_overwrite_frequency=ema_overwrite_frequency,
            loss_scale_factor=loss_scale_factor,
            gradient_accumulation_steps=gradient_accumulation_steps,
            **kwargs,
        )

    def apply(self, grads, trainable_variables=None):
        """
        Update traininable variables according to provided gradient values.

        Parameters
        ----------
        grads : list
            list of gradient tensors with 1:1 mapping to the list of variables the optimizer was built with.
        trainable_variables : list, optional
            can be provided on the first call to build the optimizer.

        """

        if len(grads) == 0:
            # It is possible that the grad is empty. In this case,
            # `apply_gradients` is a no-op.
            return

        if trainable_variables is None:
            if not self.built:
                raise ValueError(
                    "When passing `grads` without `variables`, the optimizer "
                    "must already be built on a list of variables. "
                    "Call `optimizer.build(trainable_variables)` first. "
                )
            if len(grads) != len(self._trainable_variables_indices):
                raise ValueError(
                    "When passing `grads` as a list of gradient tensors, the "
                    f"gradients must match `optimizer.variables` one-to-on. "
                    f"Received a list of {len(grads)} gradients, but the "
                    f"optimizer is tracking {len(self._trainable_variables)} "
                    "trainable variables."
                )
            trainable_variables = self._trainable_variables
        else:
            trainable_variables = list(trainable_variables)
            # Optionally build optimizer.
            if not self.built:
                with backend.name_scope(self.name, caller=self):
                    self.build(trainable_variables)
                self.built = True
            self._check_variables_are_known(trainable_variables)

        with backend.name_scope(self.name, caller=self):
            # Overwrite targeted variables directly with their gradients if
            # their `overwrite_with_gradient` is set.
            grads, trainable_variables = (
                self._overwrite_variables_directly_with_gradients(
                    grads, trainable_variables
                )
            )

            # Filter empty gradients.
            grads, trainable_variables = self._filter_empty_gradients(
                grads, trainable_variables
            )
            if len(list(grads)) == 0:
                return

            # Unscale gradients.
            scale = self.loss_scale_factor
            if scale is not None:
                grads = [g if g is None else g / scale for g in grads]

            # Apply gradient updates.
            self._backend_apply_gradients(grads, trainable_variables)

            # Apply weight decay after gradient updates
            if self.weight_decay is not None:
                for variable in trainable_variables:
                    if self._use_weight_decay(variable):
                        lr = ops.cast(self.learning_rate, variable.dtype)
                        wd = ops.cast(self.weight_decay, variable.dtype)
                        variable.assign(variable - variable * wd * lr)

            # Apply variable constraints after applying gradients.
            for variable in trainable_variables:
                if variable.constraint is not None:
                    variable.assign(variable.constraint(variable))

    def _apply_weight_decay(self, variables):
        """
        Overwrite default weight decay to return None as it is not correct for AdamW!

        Parameters
        ----------
        variables : list
            list of variables

        """

        return
